* {
  margin: 0;
  padding: 0;
}

.hbox {
  display: flex;
  flex-direction: row;
}

html,
body {
  width: 100%;
  height: 100%;
}

#wat {
  flex: 1;
  height: 100%;
  overflow: scroll;
}

#clif {
  flex: 1;
  height: 100%;
  overflow: scroll;
}

#asm {
  flex: 1;
  height: 100%;
  overflow: scroll;
}
