//! AArch64 Settings.

use crate::settings::{self, detail, Builder, Value};
use core::fmt;

// Include code generated by `cranelift/codegen/meta/src/gen_settings.rs:`. This file contains a
// public `Flags` struct with an impl for all of the settings defined in
// `cranelift/codegen/meta/src/isa/arm64.rs`.
include!(concat!(env!("OUT_DIR"), "/settings-arm64.rs"));
